#!/usr/bin/perl

# Copyright (c) International Business Machines Corporation, 2007.  All
# rights reserved. Portions of this file are supplied under license by
# Transitive Limited, which is copyright (c) Transitive Limited, as
# follows:
#
# Copyright (c) Transitive Limited 2007. All rights reserved. 
# 
# This software program, including any upgrades, patches or programs added  
# to it ("Program"), and the Materials relating to it (in particular all IPRs 
# in the foregoing) are and shall at all times remain the property of  
# Transitive Limited (English company number 04072099), which  
# reserves its rights under all applicable law in relation to that property. 
# Any use, reproduction or dealings with any part of Transitive's property which 
# has not been expressly authorised is explicitly prohibited.  Transitive confirms 
# that the Materials, technical data, information and documentation necessary for 
# maintaining and correcting the Program are confidential, and that it reserves the 
# right to prosecute any use of the Program or the Material made in breach of this  
# notice. In relation to this notice, and without limitation: 
# 
# The term "Materials" shall include all logic, logic diagrams, flowcharts,  
# orthographic representations, algorithms, routines, sub-routines, utilities,  
# modules, file structures, coding sheets, source codes listings, functional  
# specifications, specifications, user operation manuals and all other materials  
# and documents necessary to enable a reasonably skilled programmer to maintain,  
# amend and enhance the Program and whether in eye-readable or machine-readable  
# form; and 
# 
# "IPRs" shall include all registered and unregistered patents, trademarks,  
# service marks, database rights, design rights, copyright, applications for any  
# of the foregoing, and any other similar protected rights in any country including  
# the right to register or enforce the same. 

require 5.008;

use strict;
use FindBin '$RealBin';
use lib "$RealBin/../lib/perl5";
use I18N::Langinfo ();
use pAVEScripts::L10N ();


my $terminalEncoding = I18N::Langinfo::langinfo(I18N::Langinfo::CODESET());
$terminalEncoding = 'UTF-8' if $terminalEncoding =~ /ANSI_X3/;
binmode STDOUT, ":encoding($terminalEncoding)";
binmode STDERR, ":encoding($terminalEncoding)";


my $P_AVE_ERROR_TAG = "dependency_checker.pl";


sub getPrintableError()
{
  return [$!, $terminalEncoding];
}


sub print_translated
{
  my $errorCode = shift;
  my ($msg, @args) = @_;

  chomp $msg; # we don't want any newline sending to maketext

  # Watch out for tuples that specify an encoding
  my @args_copy;
  foreach (@args) {
    if (UNIVERSAL::isa($_, 'ARRAY')) {
      $_ = Encode::decode($_->[1], $_->[0]);
    }
    push @args_copy, $_;
  }

  # generate the translated output
  my $trans = pAVEScripts::L10N->get_handle()->maketext($msg, @args_copy);

  # print translated message
  print "[$P_AVE_ERROR_TAG][Error: $errorCode] $trans\n";
}

#check that root is runing the script

unless( $> == 0 ) {
  print_translated 1, "You must be root to run this script";
  exit 1;
}

# the init.d dir
my $initddir="";

if( -e "/etc/opt/p-ave/config" and -r "/etc/opt/p-ave/config" ) {
  # there is a config file, it might have a custom subject root
  open(CONFIG, "/etc/opt/p-ave/config");
  while( <CONFIG> ) {
    if( /FU_OPT_SUBJECT_WORLD_ROOT/ and not /^\s*#/ ) {
      /FU_OPT_SUBJECT_WORLD_ROOT=(.*)/;
      $initddir = "$1" . "/etc/init.d/";  
    }
  }
  close(CONFIG);
} 

if( "$initddir" eq "" ) {
  # the config file did not contain path to
  # subject world, using default
  $initddir = "/i386/etc/init.d/";
}

$initddir =~ s/\/\//\//g; # get rid of any double slash

# check if the directory exists

unless( -d $initddir ) {
  print_translated 2, "Directory [_1] does not exist.", $initddir;
  exit 1;
}

# read the list of dependencies
my $dependencies_file = "/etc/opt/p-ave/rc_dependencies"; 
my @dependencies;

if( -e $dependencies_file && -r $dependencies_file ) {
  open(DEPEN, $dependencies_file);
  @dependencies = <DEPEN>;
  close(DEPEN);
} else {
  # no list was found so lets create one
  $dependencies[$#dependencies+1] = "boot.*"; # all the boot scripts
  $dependencies[$#dependencies+1] = "acpid";
  $dependencies[$#dependencies+1] = "hwscan";
  $dependencies[$#dependencies+1] = "haldaemon";
  $dependencies[$#dependencies+1] = '\$network';
  $dependencies[$#dependencies+1] = 'network';
  $dependencies[$#dependencies+1] = '\$local_fs';
  $dependencies[$#dependencies+1] = '\$remote_fs';
  $dependencies[$#dependencies+1] = '\$named';
  $dependencies[$#dependencies+1] = '\$time';
}

#create a temp dir to process the files
my $workdir = "/tmp/p-ave_dc_$$";
mkdir("$workdir");
#to prevent someone changing the scripts while they are being edited 
chmod 0700,"$workdir";

# get the list of scripts

chdir "$initddir";
my @script_list = <*>;

# names of scripts that were changed will be put in this array
my @scripts;
# start processing scripts
foreach my $file ( @script_list ) {
  # only executables
  if( -x "$initddir/$file" && ! -d "$initddir/$file" ) {
    # open original file for reading
    unless( open(INFILE, "$initddir/$file") ) {
      print_translated 3, "Cannot open \"[_1]\" for reading: [_2].", ["$initddir/$file", $terminalEncoding], getPrintableError();
      next;
    }
    # open temporary file for writing
    unless( open(OUTFILE, ">$workdir/$file") ) {
      print_translated 4, "Cannot open \"[_1]\" for writing: [_2].", "$workdir/$file", getPrintableError();
      close INFILE;
      next;
    }
    # start reading the input file
    while( <INFILE> ) {
      my $line = $_;
      if( ($line=~/Should-Start/) or ($line=~/Required-Start/) ) {
        my $new_line = check_dependencies("$line");
        if( $new_line eq "" ) {
          # the line did not contain any black-listed services
          print OUTFILE "$line";
        } else {
          # black-listed services were found and deleted from line
          print OUTFILE "$new_line";
          $scripts[$#scripts+1]=$file;
        }
      } else {
        # line with no service names, just write it to output file
        print OUTFILE "$_";
      }
    }
    close INFILE;
    close OUTFILE;
    # restore file permissions
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size, $atime,$mtime,$ctime,$blksize,$blocks) = stat("$initddir/$file");
    chmod $mode, "$workdir/$file";
  }
}
# substitute the original file
foreach my $script(@scripts) {
  `mv "$workdir/$script" "$initddir/$script" 2>/dev/null`;
}

# No need to keep the work files anymore
unlink <$workdir/*>;
rmdir $workdir;

sub check_dependencies {
  my $the_line = shift;
  my $res = "";
  foreach my $dline (@dependencies) {
    if( $the_line =~ s/\s$dline\b// ) {
      $res = $the_line;
    }
  }
  return $res;
}

