# Copyright (c) International Business Machines Corporation, 2007.  All
# rights reserved. Portions of this file are supplied under license by
# Transitive Limited, which is copyright (c) Transitive Limited, as
# follows:
#
# Copyright (c) Transitive Limited 2007-2008. All rights reserved.
#
# This software program, including any upgrades, patches or programs added
# to it ("Program"), and the Materials relating to it (in particular all IPRs
# in the foregoing) are and shall at all times remain the property of
# Transitive Limited (English company number 04072099), which
# reserves its rights under all applicable law in relation to that property.
# Any use, reproduction or dealings with any part of Transitive's property which
# has not been expressly authorised is explicitly prohibited.  Transitive confirms
# that the Materials, technical data, information and documentation necessary for
# maintaining and correcting the Program are confidential, and that it reserves the
# right to prosecute any use of the Program or the Material made in breach of this
# notice. In relation to this notice, and without limitation:
#
# The term "Materials" shall include all logic, logic diagrams, flowcharts,
# orthographic representations, algorithms, routines, sub-routines, utilities,
# modules, file structures, coding sheets, source codes listings, functional
# specifications, specifications, user operation manuals and all other materials
# and documents necessary to enable a reasonably skilled programmer to maintain,
# amend and enhance the Program and whether in eye-readable or machine-readable
# form; and
#
# "IPRs" shall include all registered and unregistered patents, trademarks,
# service marks, database rights, design rights, copyright, applications for any
# of the foregoing, and any other similar protected rights in any country including
# the right to register or enforce the same.

package Maiken::RO;
use base 'Maiken';

=head1 NAME

  Maiken::RO - IBM p AVE installation

=cut

use strict;
use warnings;
use Cwd ();
use Fcntl ':mode';
use File::Basename ();
use File::Copy ();
use File::Find ();
use File::Path ();
use constant DATE_FORMAT => '%a %b %e %T %Z %Y';

# Set up some RO-specific configuration
sub PRODUCT_NAME          { 'p-ave'                                }
sub PRODUCT_NAME_FRIENDLY { 'PowerVM Lx86 (System p AVE)'          }
sub PRODUCT_NAME_FRIENDLY_LONG {
  # The long name and the short name are the same for RO
  'PowerVM Lx86 (System p AVE)'
}
sub IBM_REGISTER_EMAIL    { 'pave@us.ibm.com'            }
sub SW_NAME               { 'x86 World'                            }
sub SW_ROOT_PATH          { '/i386'                                }
sub SW_RPM_DIR            { '/tmp/x86rpms'                         }

sub RUNX86                { 'runx86'                               }
sub LINKX86               { 'linkx86'                              }
sub SW_PART               { '.sw_partially_installed'              }
sub SW_SUCC               { '.sw_installed_successfully'           }
sub IBM_REG_FILE          { $_[0]->CONFIG_DIR.'/activation-detail' }

sub ISO_MOUNTPOINT        { '/tmp/'.$_[0]->PRODUCT_NAME.
			      '-installer-iso-mountpoint' }

sub MAKEHOSTID { $_[0]->ROOT.'/resources/bin/RO/makeHostidFile.ppc' }
sub INIT_D { '/etc/init.d' }

# A hash of supported distributions and their details:
our %DELEGATES =
  (
   RHEL4U4  => {
		class => 'RHEL',
		name  => 'Red Hat Enterprise Linux release 4, Update 4',
		rpm   => 'RHEL4',
		subject_worlds => [qw/RHEL4U3AS_MIN RHEL4U3AS_FULL
				      RHEL4U4AS_MIN RHEL4U4AS_FULL/],
	       },
   RHEL4U5  => {
		class => 'RHEL',
		name  => 'Red Hat Enterprise Linux release 4, Update 5',
		rpm   => 'RHEL4',
		subject_worlds => [qw/RHEL4U3AS_MIN RHEL4U3AS_FULL
				      RHEL4U4AS_MIN RHEL4U4AS_FULL
				      RHEL4U5AS_MIN RHEL4U5AS_FULL/],
	       },
   RHEL4U6  => {
		class => 'RHEL',
		name  => 'Red Hat Enterprise Linux release 4, Update 6',
		rpm   => 'RHEL4',
		subject_worlds => [qw/RHEL4U3AS_MIN RHEL4U3AS_FULL
				      RHEL4U4AS_MIN RHEL4U4AS_FULL
				      RHEL4U5AS_MIN RHEL4U5AS_FULL
				      RHEL4U6AS_MIN RHEL4U6AS_FULL/],
	       },
   SLES9SP3 => {
		class => 'SLES',
		name  => 'SUSE Linux Enterprise Server release 9, Service Pack 3',
		rpm   => 'SLES9',
		subject_worlds => [qw/SLES9SP3_MIN SLES9SP3_FULL/],
	       },
   SLES9SP4 => {
		class => 'SLES',
		name  => 'SUSE Linux Enterprise Server release 9, Service Pack 4',
		rpm   => 'SLES9',
		subject_worlds => [qw/SLES9SP3_MIN SLES9SP3_FULL
				      SLES9SP4_MIN SLES9SP4_FULL/],
	       },
   SLES10   => {
		class => 'SLES',
		name  => 'SUSE Linux Enterprise Server release 10',
		rpm   => 'SLES10',
		subject_worlds => [qw/SLES10_MIN SLES10_FULL/],
	       },
   SLES10SP1 => {
		 class => 'SLES',
		 name  => 'SUSE Linux Enterprise Server release 10, Service Pack 1',
		 rpm   => 'SLES10',
		 subject_worlds => [qw/SLES10_MIN SLES10_FULL
				       SLES10SP1_MIN SLES10SP1_FULL/],
		},
   SLES10SP2 => {
                 class => 'SLES',
                 name  => 'SUSE Linux Enterprise Server release 10, Service Pack 2',
                 rpm   => 'SLES10',
                 subject_worlds => [qw/SLES10_MIN SLES10_FULL
                                       SLES10SP1_MIN SLES10SP1_FULL
                                       SLES10SP2_MIN SLES10SP2_FULL/],
                },

  );

# Subject world information
my %SUBJ_WORLDS =
  (
   RHEL4U3AS_MIN =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 3 (minimal)',
     size    => 865,
     rpmsize => 220 },
   RHEL4U3AS_FULL =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 3 (full)',
     size    => 8600,
     rpmsize => 1946 },
   RHEL4U4AS_MIN =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 4 (minimal)',
     size    => 925,
     rpmsize => 236 },
   RHEL4U4AS_FULL =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 4 (full)',
     size    => 7680,
     rpmsize => 1965 },
   RHEL4U5AS_MIN =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 5 (minimal)',
     size    => 967,
     rpmsize => 251},
   RHEL4U5AS_FULL =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 5 (full)',
     size    => 7783,
     rpmsize => 2048},		 
   RHEL4U6AS_MIN =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 6 (minimal)',
     size    => 903,
     rpmsize => 278},
   RHEL4U6AS_FULL =>
   { title   => 'Red Hat Enterprise Linux AS release 4, Update 6 (full)',
     size    => 7400,
     rpmsize => 2210},		 
   SLES9SP3_MIN =>
   { title   => 'SUSE Linux Enterprise Server release 9, Service Pack 3 (minimal)',
     size    => 430,
     rpmsize => 119 },
   SLES9SP3_FULL =>
   { title   => 'SUSE Linux Enterprise Server release 9, Service Pack 3 (full)',
     size    => 2662,
    rpmsize => 773 },
   SLES9SP4_MIN =>
   { title   => 'SUSE Linux Enterprise Server release 9, Service Pack 4 (minimal)',
     size    => 500,
     rpmsize => 140 },
   SLES9SP4_FULL =>
   { title   => 'SUSE Linux Enterprise Server release 9, Service Pack 4 (full)',
     size    => 3000,
    rpmsize => 880 },
   SLES10_MIN =>
   { title   => 'SUSE Linux Enterprise Server release 10 (minimal)',
     size    => 642,
     rpmsize => 192 },
   SLES10_FULL =>
   { title   => 'SUSE Linux Enterprise Server release 10 (full)',
     size    => 2560,
     rpmsize => 740 },
   SLES10SP1_MIN =>
   { title   => 'SUSE Linux Enterprise Server release 10, Service Pack 1 (minimal)',
     size    => 686,
     rpmsize => 206 },
   SLES10SP1_FULL =>
   { title   => 'SUSE Linux Enterprise Server release 10, Service Pack 1 (full)',
     size    => 2765,
     rpmsize => 797 },
   SLES10SP2_MIN =>
   { title   => 'SUSE Linux Enterprise Server release 10, Service Pack 2 (minimal)',
     size    => 780,
     rpmsize => 212 },
   SLES10SP2_FULL =>
   { title   => 'SUSE Linux Enterprise Server release 10, Service Pack 2 (full)',
     size    => 3100,
     rpmsize => 869 },
  );

=item B<initialize_delegate>

Delegate to a class determined by determine_linux_distribution.

=cut

sub initialize_delegate {
  my ($class) = @_;
  my $self = bless({}, $class); # temporary $self

  my $distro = $self->determine_linux_distribution;

  # Nice message if distro isn't supported
  if (!exists $DELEGATES{$distro}) {
    $self->log("Your OS distribution ([_1]) is unsupported. Only the following are supported:", $distro);
    foreach my $details (sort {$a->{name} cmp $b->{name}} values %DELEGATES) {
      $self->log("- [_1]", $details->{name} );
    }
    exit 1;
  }

  $self->log("Host OS detected as [_1]", $DELEGATES{$distro}{name});
  $self->{distribution} = $distro;
  $self->{distro_details} = $DELEGATES{$distro};

  # Delegate to the appropriate class
  $class .= "::$DELEGATES{$distro}{class}";
  eval qq{ require $class; };
  if ($@) {
    $self->croak("Error loading [_1]: [_2]", $class, $@);
  }

  # Now we've loaded the class, rebless $self into it
  bless($self, $class);
  return $self;
}

=item B<set_up_environment>

  umask 022

=cut

sub set_up_environment {
  my ($self) = @_;
  umask 022;
  $ENV{PATH} = '/sbin:/usr/sbin:/usr/local/bin:/usr/bin:/bin';
  if (not -d $self->CONFIG_DIR) {
    File::Path::mkpath($self->CONFIG_DIR);
  }
}

=item B<check_environment>

  Are we on a ppc system with the appropriate /proc stuff?
  Check that we have required RPMs installed.
  set_default_config_common
  Is the translator RPM installed?

=cut

sub check_environment {
  my ($self) = @_;
  my %unsupported_models;
  @unsupported_models { qw/ 
      9406-B30 9406-B35 9406-B40 9406-B45 9406-B50 9406-B60 9406-B70 9406-D35 9406-D45
      9406-D50 9406-D60 9406-D70 9406-D80 9406-E35 9406-E45 9406-E50 9406-E60 9406-E70
      9406-E80 9406-E90 9406-E95 9406-F35 9406-F45 9406-F50 9406-F60 9406-F70 9406-F80
      9406-F90 9406-F95 9406-F97 9406-SB1 9406-S10 9406-S20 9406-S30 9406-S40 9406-000
      9406-170 9406-250 9406-270 9406-30S 9406-300 9406-310 9406-320 9406-50S 9406-500
      9406-510 9406-520 9406-53S 9406-530 9406-550 9406-570 9406-590 9406-595 9406-600
      9406-620 9406-640 9406-650 9406-720 9406-730 9406-740 9406-800 9406-810 9406-820
      9406-825 9406-830 9406-840 9406-870 9406-890 9411-100 9405-520 9404-B10 9404-B20
      9404-C10 9404-C20 9404-C25 9404-D00 9404-D10 9404-D20 9404-D25 9404-E00 9404-E10
      9404-E20 9404-E25 9404-F00 9404-F10 9404-F20 9404-F25 9404-SB1 9404-S10 9404-S20
      9404-S30 9404-S40 9404-135 9404-140 9404-170 9404-200 9404-30S 9404-300 9404-310
      9404-320 9404-40S 9404-400 9404-50S 9404-500 9404-510 9404-53S 9404-530 9404-600
      9404-620 9404-640 9404-650 9404-720 9404-730 9404-740 9404-800 9404-810 9404-820
      9404-825 9404-830 9404-840 9404-870 9404-890 9402-C04 9402-C06 9402-D00 9402-D02
      9402-D04 9402-D06 9402-E00 9402-E02 9402-E04 9402-E06 9402-F02 9402-F04 9402-F06
      9402-S10 9402-S20 9402-Y10 9402-100 9402-170 9402-20S 9402-200 9402-236 9402-40S
      9402-400 9402-436 9402-600 9402-620 9402-720 9402-730 9402-740 9402-800 9402-810
      9402-820 9402-830 9402-840 9401-P01 9401-P02 9401-P03 9401-P10 9401-P20 9401-10S
      9401-150 / } = ();

  # Is it running under translation or on a non-PPC machine?
  if ($self->run_command('uname', '-m') ne "ppc64\n") {
    $self->croak("This installer must be run on an IBM System p platform");
  }

  # Look for IBM system information
  if (-d '/proc/device-tree') {
    $self->{config}{model} = $self->slurp('/proc/device-tree/model');
    chomp $self->{config}{model};
    $self->{config}{model} =~ s/\0$//;
    $self->{config}{serial} = $self->slurp('/proc/device-tree/system-id');
    chomp $self->{config}{serial};
    $self->{config}{serial} =~ s/\0$//;
    $self->{config}{model} =~ /IBM,(.*)/;
    my $model_number = $1;
    if( not defined($model_number) or exists( $unsupported_models{$model_number} ) ) {
      $self->croak("This machine model is not supported.\nPlease check the system requirements in the PowerVM Lx86 (System p AVE) Administration Guide.");
    }
  }
  else {
    $self->croak("Either this is not an IBM system or /proc is not mounted");
  }

  # Common config
  $self->set_default_config_common;

  # Information about the QT RPM
  $self->{config}{rpm_installed} =
    $self->get_installed_rpm_full_name($self->PRODUCT_NAME);
  unless ($self->{quitting}) {
    $self->{config}{versions_available} = $self->versions_available;
  }

  $self->check_sw_installation;

  # sw_rpm_dir - set from environment or use default
  if ($ENV{SW_RPM_DIR}) {
    $self->log("WARNING: using non-standard RPM staging directory [_1].",
	       $ENV{SW_RPM_DIR});
  }
  $self->{config}{sw_rpm_dir} = $ENV{SW_RPM_DIR} || $self->SW_RPM_DIR;
}

=item B<check_sw_installation>

Is the SW installed? What Distro is it?

=cut

sub check_sw_installation {
  my ($self) = @_;
  if (-f "$self->{config}{sw_root_path}/etc/".$self->SW_SUCC) {
    $self->{config}{sw_installed} =
      $self->determine_linux_distribution($self->{config}{sw_root_path});
  }
  else {
    $self->{config}{sw_installed} = undef;
  }
}

=item B<get_initial_config_interactively>

  get_registration_info_interactively

=cut

sub get_initial_config_interactively {
  my ($self) = @_;
  $self->license_agreement_interactively('utf16', 'LA', 'LI') or return;
  $self->get_registration_info_interactively or return;
  return 1;
}

=item B<get_install_config_interactively>

Ask what things the user wants to install.

Get the paths for various things.

=cut

sub get_install_config_interactively {
  my ($self) = @_;
  my $action = $self->prompt_user_for_install;
  return 0 if $action eq 'return';

  # Remember what the user wants to install
  $self->{config}{install_qt} = ($action =~ /qt/);
  $self->{config}{install_sw} = ($action =~ /sw/);

  # Can't install the SW without QT
  if ($self->{config}{install_sw} and
      not ($self->{config}{install_qt} or $self->{config}{rpm_installed})) {
    $self->log;
    $self->log("Cannot install [_1] when [_2] is not installed.",
	       $self->SW_NAME, $self->PRODUCT_NAME);
    return 0;
  }

  # Is QT already installed?
  if ($self->{config}{install_qt} && $self->{config}{rpm_installed}) {
    $self->log("Cannot install [_1] because [_2] is already installed.",
	       $self->PRODUCT_NAME, $self->{config}{rpm_installed});
    return 0;
  }

  do {
    $self->check_sw_installation; # Since we may be looking in a
                                  # different place now
    # Is the SW already installed?
    if ($self->{config}{install_sw} && $self->{config}{sw_installed}) {
      $self->log("Cannot install [_1] because it ([_2]) is already installed", 
		 $self->SW_NAME, $self->{config}{sw_installed});
      return 0;
    }
    $self->get_config_common_interactively;
  }
  until ($self->check_sw_and_qt_paths_differ);

  # Ban /mnt and /media from being filesystems of their own
  if ($self->{config}{install_sw}) {
    foreach my $dir (qw|/mnt /media|) {
      if (-d $dir and (stat $dir)[0] != (stat '/')[0]) {
	$self->log("Unable to install [_1] because [_2] is a mounted filesystem.\nPlease unmount it before installing [_1].",
		   $self->SW_NAME, $dir);
	return 0;
      }
    }
  }

  # Ask the user which version should be installed
  if ($self->{config}{install_qt}) {
    $self->{config}{chosen_rpm_file} = $self->version_menu;
    return 0 if $self->{config}{chosen_rpm_file} eq 'return';
    # Check the RPM can be installed and bomb out if it can't.
    if (not $self->install_rpm($self->{config}{chosen_rpm_file}, test => 1)) {
      $self->log;
      $self->log("RPM installation test failed.");
      $self->log($self->{rpm_output});
      return 0;
    }
  }

  if ($self->{config}{install_sw}) {
    $self->get_subject_world_config_interactively or return 0;
  }

  # Final confirmation before installing the subject world
  $self->log;
  $self->log("About to proceed with the following:");
  $self->log;
  if ($self->{config}{install_qt}) {
    $self->log("- Install [_1] to [_2]",
	       $self->PRODUCT_NAME,
	       [$self->{config}{qt_root_path}, 'user']);
    $self->log("  - Log files will be stored in [_1]",
	       [$self->{config}{logfile_path}, 'user']);
  }
  if ($self->{config}{install_sw}) {
    $self->log("- Install [_1] to [_2]",
	       $self->SW_NAME,
	       [$self->{config}{sw_root_path}, 'user']);
    $self->log("  - Selected distribution is [_1]",
	       $SUBJ_WORLDS{$self->{config}{sw_distro}}{title});
    if (defined $self->{config}{sw_homedir}
	and $self->{config}{sw_homedir} ne 'skip') {
      $self->log("  - Home directories are in [_1]",
		 [$self->{config}{sw_homedir}, 'user']);
    }
    else {
      $self->log("  - Home directories will not be shared");
    }
  }
  my $answer = $self->menu([
			    [cont => "Continue."],
			    [edit => "Change settings."],
			    [return => "Go back to the Main Menu."],
			   ]);
  return 0 if $answer eq 'return';
  # Do it again?
  return $self->get_install_config_interactively if $answer eq 'edit';

  return 1;
}


=item B<check_sw_and_qt_paths_differ>

Checks that we're not trying to install p-ave and the x86world
to the same directory.

Also check that p-ave will not be installed to a subdirectory of x86
World.

=cut

sub check_sw_and_qt_paths_differ {
  my ($self) = @_;

  # Are both components going to end up installed on this machine?
  if ( ($self->{config}{install_qt} || $self->{config}{rpm_installed}) &&
       ($self->{config}{install_sw} || $self->{config}{sw_installed}) ) {

    if ($self->{config}{sw_root_path} eq $self->{config}{qt_root_path} or
	$self->{config}{qt_root_path} =~ m|^\Q$self->{config}{sw_root_path}\E/|) {
      $self->log("You cannot install [_1] and [_2] into the same directory." ,
		              $self->PRODUCT_NAME, $self->SW_NAME);
      $self->log("Please re-enter your installation paths.") ;

      return 0 ;
    }
  }

  return 1 ;
}

=item B<prompt_user_for_install>

Submenu for 'install'.

=cut

sub prompt_user_for_install {
  my ($self) = @_;
  my $options = [
		 [qt_sw => ["Install [_1] + [_2].",
			    $self->PRODUCT_NAME, $self->SW_NAME]],
		 [qt    => ["Install [_1].", $self->PRODUCT_NAME]],
		 [sw    => ["Install [_1].", $self->SW_NAME]],
		 [return => "Go back to the Main Menu."],
		 [quit   => "Quit."],
		];
  my $action = $self->menu($options);
  $self->quit if $action eq 'quit';
  return $action;
}

=item B<get_subject_world_config_interactively>

To install the subject world we need some extra information from the
user.

=cut

sub get_subject_world_config_interactively {
  my ($self) = @_;

  # Offer to remove anything that doesn't look like a fully-installed
  # SW.
  if (-d $self->{config}{sw_root_path}) {
    if ($self->is_on_remote_filesystem($self->{config}{sw_root_path})) {
      $self->log;
      $self->log("[_1] appears to be on a remote filesystem. Installing to remote filesystems is unsupported.", [$self->{config}{sw_root_path}, 'user']);
      return 0;
    }

    my @files_in_dir = $self->read_directory($self->{config}{sw_root_path});
    # If the directory is empty, then just use it.
    if (@files_in_dir) {
      if (-f "$self->{config}{sw_root_path}/etc/".$self->SW_PART) {
	      $self->log("Found partially installed [_1] at [_2].",
		      $self->SW_NAME, [$self->{config}{sw_root_path}, 'user']);
      }
      elsif (-f "$self->{config}{sw_root_path}/etc/".$self->SW_SUCC) {
        $self->log("Found installed [_1] at [_2].",
          $self->SW_NAME, [$self->{config}{sw_root_path}, 'user']);
      }
      else {
	     $self->log("Directory [_1] already exists but doesn't look like an installation of [_2]", [$self->{config}{sw_root_path}, 'user'],
		     $self->SW_NAME);
      }

      # Empty the directory. We don't delete it in case it's a mount point.
      if ($self->ask(q => ["Empty the directory [_1]?",
			   [$self->{config}{sw_root_path}, 'user']
			   ], type => 'bool',
		     default => 0
		    )) {
	my $mounted = $self->directory_has_mounted_filesystems
	  ($self->{config}{sw_root_path});
	if ($mounted) {
	  $self->log("The directory [_1] contains mounted filesystems ([_2]). Please unmount them before retrying this operation.", [$self->{config}{sw_root_path}, 'user'], $mounted);
	  return 0;
	}
	$self->log("Emptying directory...");
	foreach (@files_in_dir) {
	  File::Path::rmtree("$self->{config}{sw_root_path}/$_") or
	      $self->croak("Unable to remove [_1]",
			   ["$self->{config}{sw_root_path}/$_", 'user']);
	}
      }
      else {
	$self->log("Aborting installation.");
 	return 0;
      }
    }
  }
  else {
    # Create the SW directory
    push @{$self->{cleanup_dirs}},
      File::Path::mkpath($self->{config}{sw_root_path});
    if ($self->is_on_remote_filesystem($self->{config}{sw_root_path})) {
      $self->log;
      $self->log("[_1] appears to be on a remote filesystem. Installing to remote filesystems is unsupported.", [$self->{config}{sw_root_path}, 'user']);
      return 0;
    }
  }


  # Give the user a choice of distribution for the subject world.
  my $options = [];
  foreach my $sw (sort @{$self->{distro_details}{subject_worlds}}) {
    push @$options, [$sw, $SUBJ_WORLDS{$sw}{title}];
  }
  if (@$options > 1) {
    $self->log("Which Linux/x86 OS distribution will you be installing for your [_1]?",
	       $self->SW_NAME);
    $self->{config}{sw_distro} =
      $self->menu($options,
		  "Please select a distribution from the list above:");
  }
  else {
    # If we only have one, just choose it automatically
    $self->{config}{sw_distro} = $options->[0][0];
  }

  $self->{config}{sw_homedir} = undef; # ask the question at least once
  while (not defined $self->{config}{sw_homedir}) {
    $self->{config}{sw_homedir} =
      $self->ask(q => ["Where are the home directories located on this system (type 'skip' if you don't want the [_1] to share your home directories)?", $self->SW_NAME],
		 type => 'path', default => '/home', skip => 1,
		 must_exist => 1);
    # Tidy it up
    if ($self->{config}{sw_homedir} ne 'skip') {
      $self->{config}{sw_homedir} = Cwd::abs_path($self->{config}{sw_homedir});
      # Remove trailing slash
      $self->{config}{sw_homedir} =~ s|/$||
	if defined $self->{config}{sw_homedir};
    }
  }

  # Create the temporary RPM directory
  $self->log;
  $self->log(<<EOT, $self->SW_NAME, $self->PRODUCT_NAME, $SUBJ_WORLDS{$self->{config}{sw_distro}}{title});
To create the [_1] the [_2] installer will need the appropriate set of x86
RPMs. This set of RPMs is generated from the [_3] x86 install media
(ISO or CD/DVD-ROM).

EOT
  $self->{config}{rpms_still_to_get} = $self->get_rpm_list_as_hashref;
  # Regex_rpms are the special RPM names that are regexes, delimited by //
  my @regex_rpms = grep s|^/(.*)/$|$1|,
    keys %{$self->{config}{rpms_still_to_get}};
  $self->{config}{regex_rpms} = \@regex_rpms;
  $self->{config}{real_names_for_regexes} = {};
  $self->{config}{total_rpms_required} =
    scalar keys %{$self->{config}{rpms_still_to_get}};
  my $base_rpms = $self->get_rpm_list_as_hashref('base');
  $self->{config}{total_base_rpms_required} = scalar keys %$base_rpms;
  $self->{config}{rpms_found} = 0;

  if (-d $self->{config}{sw_rpm_dir}) {
    # Prompt the user if I'm about to remove a non-standard RPM directory
    if ($self->{config}{sw_rpm_dir} ne $self->SW_RPM_DIR) {
      $self->ask(q => ["Non-standard RPM staging directory [_1] already ".
		       "exists! OK to remove it?",
		       $self->{config}{sw_rpm_dir}],
		 type => 'bool', default => 1) or return 0;
    }
    File::Path::rmtree($self->{config}{sw_rpm_dir});
  }

  # Recreate it
  File::Path::mkpath($self->{config}{sw_rpm_dir});

  # Check the disk space is OK
  $self->check_disk_space_in_sw_rpm_dir or return 0;
  $self->check_disk_space_in_sw_root_dir or return 0;

  $self->populate_sw_rpm_dir or return 0;
  $self->log("All RPMs found successfully.");

  return 1;
}

=item B<install>

User can choose to install QT, SW or both.

=cut

sub install {
  my ($self) = @_;

  $self->stop_qt;
  if ($self->{config}{install_qt}) {
    $self->install_qt or return 0;
  }
  if ($self->{config}{install_sw}) {
    if (! -d $self->{config}{sw_rpm_dir}) {
      $self->croak("Failed to install subject world: [_1] does not exist",
		   $self->{config}{sw_rpm_dir});
    }
    $self->install_sw or return 0;
  }
  else {
    # If we're not installing the SW we should start the daemon back
    # up here.
    $self->start_qt;
  }

  $self->check_environment;
  if ($self->{config}{rpm_installed} and $self->{config}{sw_installed}) {
    $self->bind_qt_and_sw or return 0;
  }

  return 1;
}

=item B<install_qt>

Install the QT RPM and perform some postprocessing.

Also install the installer into the qt_root_path. It's done here
because we don't know the qt_root_path until this point.

Similarly, linkx86 is created at this stage because it needs access to
the installer libraries.

=cut

sub install_qt {
  my ($self) = @_;

  # Install the installer into the specified qt_root_path
  my $installer_path = "$self->{config}{qt_root_path}/installer";
  $self->install_self_into($installer_path);

  # Symlink the doc/ directory
  symlink "$installer_path/resources/doc/RO",
    "$self->{config}{qt_root_path}/doc";

  # RPM
  $self->install_rpm($self->{config}{chosen_rpm_file},
		     relocate => [$self->QT_ROOT_PATH,
				  $self->{config}{qt_root_path}]) or
    $self->croak("RPM installation failed.");

  # Config file
  $self->write_config_file;

  # Log directory

  # Check that the parent of the logfile directory is read-only
  my $log_path        = $self->{config}{logfile_path};
  my $log_path_parent = File::Basename::dirname($log_path);
  if (-d $log_path_parent) {
    my $mode = (stat $log_path_parent)[2];
    if ( !(-O $log_path_parent)
      || ($mode & S_IWGRP)
      || ($mode & S_IWOTH))
    {
      $self->log(
        "WARNING: The directory [_1] is writable by non-root users. "
       ."This is insecure and should be made read-only.",
       $log_path_parent
      );
    }
  }

  umask 000;

  # Create any parent directories read-only, then chmod the log path
  File::Path::mkpath($log_path, undef, 0755);
  chmod(01777, $log_path);
  umask 022;

  # Symlink linkx86 from the installed installer path
  my $linkx86_symlink = $self->BIN_DIR.'/'.$self->LINKX86;
  my $linkx86_source = "$self->{config}{qt_root_path}/installer/resources/bin/RO/".$self->LINKX86;
  if (-l $linkx86_symlink or -e $linkx86_symlink) {
    unlink $linkx86_symlink or
      $self->croak("Failed to remove [_1]", $linkx86_symlink);
  }
  symlink($linkx86_source, $linkx86_symlink) or
	    $self->croak("Failed to create symlink [_1]", $linkx86_symlink);
  if (not -x $linkx86_source) {
    chmod 0755, $linkx86_source;
  }

  # Success! Add to transactions
  $self->log;
  $self->log("[_1] installed successfully.", $self->PRODUCT_NAME);
  $self->log;
  $self->add_transaction("Installed [_1]", $self->PRODUCT_NAME_FRIENDLY);
  $self->add_transaction("Installed installer into [_1]", [$installer_path, 'user']);
}

=item B<install_sw>

Install a Subject World.

=cut

sub install_sw {
  my ($self) = @_;

  $self->log;
  $self->log("Installing [_1]...", $self->SW_NAME);
  $self->log;

  $self->write_config_file;
  $self->initialize_sw_directory;
  $self->install_base_rpms_natively;
  $self->prepare_sw_for_translated_install;
  $self->install_rpms_translated;
  $self->verify_sw_installation;
  $self->setup_sw;
  $self->setup_sw_interactive;
  $self->finalize_sw_installation;
  $self->cleanup_after_sw_installation;

  $self->log;
  $self->log("[_1] installed successfully.", $self->SW_NAME);
  $self->log;
  $self->add_transaction("Installed [_1] ([_2])", $self->SW_NAME,
			 $SUBJ_WORLDS{$self->{config}{sw_distro}}{title});
}

=item B<initialize_sw_directory>

Create some directories and empty files.

Copy /etc/passwd and friends from the host.

=cut

sub initialize_sw_directory {
  my ($self) = @_;

  # Make directories
  foreach (qw|/etc /var/lock/rpm /var/lib/rpm|) {
    File::Path::mkpath("$self->{config}{sw_root_path}$_");
  }
  # Make empty files
  foreach ('/etc/'.$self->SW_PART, '/var/lock/rpm/transaction') {
    $self->write_file("$self->{config}{sw_root_path}$_");
  }

  $self->replicate_passwd_and_friends($self->{config}{sw_root_path});
}

=item B<install_base_rpms_natively>

Install the 'base' RPMs natively using --root.

=cut

sub install_base_rpms_natively {
  my ($self) = @_;

  $self->run_command('rpm', '--root', $self->{config}{sw_root_path},
		     '--initdb');
  if ($self->{exit}) {
    $self->croak("Failure initializing RPM database");
  }

  # Install the base RPMs natively
  $self->log("Step 1/2: Installing base RPMs...");
  $self->install_rpm_list(list => 'base', translated => 0, base_only => 1);
  $self->log;
}

=item B<install_rpm_list>

Install an RPM list. Takes the following options:

  list (compulsory): the name of the list (eg. 'base')
  translated:        if true, runs slightly different commands in order to
                     install the RPM list translated
  installed:         a hashref of RPMs currently installed that can be
                     used as an accumulator
  base_only:         Count only base RPMs when calculating percentage? Boolean.

Returns without doing anything if the file doesn't exist.

=cut

sub install_rpm_list {
  my ($self, %o) = @_;

  my $rpmlist_dir = $self->get_rpm_list_dir;
  return if not -f "$rpmlist_dir/$o{list}";

  my @rpm_lines = $self->slurp("$rpmlist_dir/$o{list}");

  # Restore from reference or create fresh
  my $installed = ref($o{installed}) ? $o{installed} : {};

  #
  # How many RPMs are we installing? It depends on whether we're
  # installing all RPMs, or just the base ones.
  #
  my $num_rpms_required;
  if ($o{base_only}) {
    $num_rpms_required = $self->{config}{total_base_rpms_required};
  }
  else {
    $num_rpms_required = $self->{config}{total_rpms_required};
  }

  if (@rpm_lines > 1 and $o{list} eq 'base') {
    $self->percent(0, $num_rpms_required)
  }

  $self->{rpm_start_time} ||= time();
  foreach my $rpm_line (@rpm_lines) {
    my @rpms = split /\s+/, $rpm_line;
    next unless @rpms;
    # Substitute regexes for their real names
    @rpms = map {
      m|^/(.*)/$| ? $self->{config}{real_names_for_regexes}{$1} : $_;
    } @rpms;
    # Add path info
    # Bear in mind that rpm commands may be in this list.
    @rpms = map {
      /\.rpm$/ ? "$self->{config}{sw_rpm_dir}/$_" : $_
    } @rpms;

    # This bit is different native/translated
    if (not $o{translated}) {
      my @command = ('rpm', '--root', $self->{config}{sw_root_path},
		     '--ignorearch', '--replacepkgs', '--noscripts',
		     '--nosignature', '--nodigest', '-iv', @rpms, '2>&1');
      $self->log_only_to_file($self->run_command(@command));
      if ($self->{exit}) {
	$self->croak("Failure installing [_1]", "@rpms"); # quotes necessary
      }
    }
    else {
      my @cmd = ('/bin/rpm', '--nosignature', '--nodigest',
		 '--replacefiles', '--replacepkgs', '-iv', @rpms);
      push @cmd, '2>&1';

      # RPMs beginning 'filesystem-' need the /proc moving out of the way
      if ($rpm_line =~ /^filesystem-/) {
	rename "$self->{config}{sw_root_path}/proc",
	  "$self->{config}{sw_root_path}/proc.bak";
	$self->log_only_to_file($self->run_translated(@cmd));
	rename "$self->{config}{sw_root_path}/proc.bak",
	  "$self->{config}{sw_root_path}/proc";
      }
      else {
	$self->log_only_to_file($self->run_translated(@cmd));
      }

      if ($o{translated}) {
        $self->check_login_defs($rpm_line);
      }
    }

    my @real_rpms = grep { /\.rpm$/ } @rpms;
    @$installed{@real_rpms} = ();
    if (@rpm_lines > 1) {
      $self->percent(scalar keys %$installed,  $num_rpms_required);
    }
  }
}

=item B<prepare_sw_for_translated_install>

Chdir. Escape some stuff. Make some files. Start QT.

=cut

sub prepare_sw_for_translated_install {
  my ($self) = @_;

  # Change min uid/gid in /etc/login.defs before installing anything translated
  $self->modify_login_defs;

  # The /var/lib/rpm/ can cause problems. Throw its contents away.
  File::Path::rmtree("$self->{config}{sw_root_path}/var/lib/rpm");
  File::Path::mkpath("$self->{config}{sw_root_path}/var/lib/rpm");

  # Get inside the SW before doing translated operations
  chdir $self->{config}{sw_root_path} or $self->croak("Couldn't chdir");

  #File::Path::rmtree("$self->{config}{sw_root_path}/proc");

  # Escape some things necessary for the next stage.
  $self->escape(qw{/dev /sys /selinux /mnt /media /var/yp/binding});

  # Copy the /etc/hosts file from the native system
  $self->copy_preserve("/etc/hosts", "$self->{config}{sw_root_path}/etc/hosts");

  # Copy the /etc/resolv.conf file from the native system
  $self->copy_preserve("/etc/resolv.conf", "$self->{config}{sw_root_path}/etc/resolv.conf");

  # Create stub mtab file
  $self->write_file("$self->{config}{sw_root_path}/etc/mtab");

  # Create hostid file with the makeHostid program
  if (not -x $self->MAKEHOSTID) {
    chmod 0755, $self->MAKEHOSTID;
  }
  $self->run_command($self->MAKEHOSTID,
		     "$self->{config}{sw_root_path}/etc/hostid");
  if ($self->{exit}) {
    $self->croak("[_1] failed", $self->MAKEHOSTID);
  }

  # Symlink the RPM directory into the SW
  $self->escape($self->{config}{sw_rpm_dir});

  # Start QT up ready for the translated stage
  $self->start_qt;
}

=item B<install_rpms_translated>

Install all RPMs from the 'base' and 'apps' lists now under translation.

=cut

sub install_rpms_translated {
  my ($self) = @_;

  # Reinitialize the database
  $self->run_translated('/bin/rpm', '--initdb');
  if ($self->{exit}) {
    $self->croak("Failure initializing RPM database");
  }
  $self->log("Step 2/2: Installing [_1]...", $self->SW_NAME);
  my $installed = {};
  delete $self->{rpm_start_time}; # as don't want to inherit nontranslated timings
  foreach my $file (qw/base apps/) {
    $self->install_rpm_list(list => $file, translated => 1,
			    installed => $installed);
  }
  $self->log;
}

=item B<verify_sw_installation>

Get the number of installed RPMs. Croak if it is not the same as the
the number of RPMs we were expecting to install.

=cut

sub verify_sw_installation {
  my ($self) = @_;
  my $actually_installed = $self->run_translated('/bin/rpm', '-qa');
  my @n_installed = split "\n", $actually_installed;
  if (@n_installed != $self->{config}{total_rpms_required}) {
    $self->log_only_to_file("Missing RPMs:\n".$self->get_missing_rpms(\@n_installed));
    $self->croak("Failed to install all required rpms");
  }
}

=item B<get_missing_rpms>

Given a reference to an array of installed RPM names, find the RPMs
that should have been installed but were, and return them as a string.

=cut

sub get_missing_rpms {
  my ($self, $installed) = @_;
  my %expected = %{$self->get_rpm_list_as_hashref};

  # Check each of the expected ones and make sure we installed one like it
  my @missing;
  foreach my $rpm (keys %expected) {
    $rpm =~ s/\\?\.\w+\\?\.rpm//; # remove the bit that rpm -q doesn't return
    my $regex = "^\Q$rpm";

    # Unless it's a regex-rpm
    if ($rpm =~ m|^/(.*)/$|) {
      $regex = $1;
    }

    push @missing, $rpm unless grep /$regex/, @$installed;
  }

  return join("\n", @missing)."\n";
}

=item B<setup_sw>

Set up various escape paths.

Bind /home.

Create mtab and fstab

=cut

sub setup_sw {
  my ($self) = @_;
  if (not $self->{config}{sw_root_path}) {
    $self->croak("sw_root_path is not set!");
  }
  $self->escape(qw{/dev /sys /selinux /mnt /media
		   /var/run/.nscd_socket /etc/localtime});

  # Escape logfile path and home directories
  $self->escape($self->{config}{logfile_path});
  if ($self->{config}{sw_homedir} ne 'skip') {
    $self->escape($self->{config}{sw_homedir}, escape_anyway => 1);
  }

  # Copy the /etc/hosts file from the native system
  $self->copy_preserve("/etc/hosts", "$self->{config}{sw_root_path}/etc/hosts");

  # Copy the /etc/resolv.conf file from the native system
  $self->copy_preserve("/etc/resolv.conf", "$self->{config}{sw_root_path}/etc/resolv.conf");

  # Create stub utmp file
  $self->write_file("$self->{config}{sw_root_path}/var/run/utmp");

  # Create stub wtmp file
  $self->write_file("$self->{config}{sw_root_path}/var/log/wtmp");

  # Generate fstab
  my @fstab = $self->slurp('/etc/fstab');
  my @sw_fstab;
  foreach (@fstab) {
    my @f = split;
    next unless @f >= 3;
    if ($f[1] eq '/' or $f[2] =~ /^(proc|sysfs|swap|devpts|tmpfs)$/) {
      push @sw_fstab, $_;
    }
  }
  $self->write_file("$self->{config}{sw_root_path}/etc/fstab", \@sw_fstab);
}

=item B<setup_sw_interactive>

Those parts of setup_sw that require user input.

This function is empty now.

=cut

sub setup_sw_interactive {
  my ($self) = @_;
}

=item B<finalize_sw_installation>

Remove any existing .rpmorig or .rpmnew files from the SW.

Update the flag to say SW is successfully installed.

=cut

sub finalize_sw_installation {
  my ($self) = @_;
  $self->log("Finalizing [_1] installation. Please wait...", $self->SW_NAME);

  # Ensure that any files named *.rpmorig or *.rpmnew are removed after the
  # install
  my $find = sub {
    if (/\.rpm(new|orig)$/) {
      unlink($File::Find::name) or
	$self->croak("Cannot perform post install remove of [_1]",
		     $File::Find::name);
    }
  };
  File::Find::find($find, $self->{config}{sw_root_path});

  # Finally move the flag to indicate the Subject World is installed
  # successfully.
  rename "$self->{config}{sw_root_path}/etc/".$self->SW_PART,
    "$self->{config}{sw_root_path}/etc/".$self->SW_SUCC;

  # Generate mtab
  # NOTE: this needs to go here, after we stamp the x86 World with
  #       the .sw_installed_successfully file, since the *-world-sync
  #       will exit quietly if it's not there.
  my $world_sync = "/usr/sbin/".$self->PRODUCT_NAME."-world-sync";
  $self->run_command($world_sync, "force_sync_mtab");
  if ($self->{exit}) {
    $self->croak("[_1] failed", $world_sync);
  }
}

=item B<cleanup_after_sw_installation>

Remove the RPM staging directory.

=cut

sub cleanup_after_sw_installation {
  my ($self) = @_;
  $self->log("Removing [_1]...", $self->{config}{sw_rpm_dir});
  File::Path::rmtree($self->{config}{sw_rpm_dir});
  unlink "$self->{config}{sw_root_path}$self->{config}{sw_rpm_dir}";
}

=item B<bind_qt_and_sw>

If the QT or SW has changed, we need to recreate the runx86 script.

=cut

sub bind_qt_and_sw {
  use utf8; # make sure the string literal is in utf8 - not strictly
            # necessary here but you never know, eh?
  my ($self) = @_;
  $self->restart_qt;
  $self->log;
  $self->log("[_1] and [_2] are now installed.",
	     $self->PRODUCT_NAME, $self->SW_NAME);
  $self->log;
  $self->log("Binding them together and creating [_1] script.", $self->RUNX86);

  # Create the script
  my $qt_start = "$self->{config}{qt_root_path}/bin/".$self->PRODUCT_NAME;
  my $runx86_script = qq{#!/bin/bash
# Copyright (c) International Business Machines Corporation, 2007.  All
# rights reserved. Portions of this file are supplied under license by
# Transitive Limited, which is copyright (c) Transitive Limited, as
# follows:
#
# Copyright (c) Transitive Limited 2007-2008. All rights reserved.
#
# This software program, including any upgrades, patches or programs added
# to it ("Program"), and the Materials relating to it (in particular all IPRs
# in the foregoing) are and shall at all times remain the property of
# Transitive Limited (English company number 04072099), which
# reserves its rights under all applicable law in relation to that property.
# Any use, reproduction or dealings with any part of Transitive's property which
# has not been expressly authorised is explicitly prohibited.  Transitive confirms
# that the Materials, technical data, information and documentation necessary for
# maintaining and correcting the Program are confidential, and that it reserves the
# right to prosecute any use of the Program or the Material made in breach of this
# notice. In relation to this notice, and without limitation:
#
# The term "Materials" shall include all logic, logic diagrams, flowcharts,
# orthographic representations, algorithms, routines, sub-routines, utilities,
# modules, file structures, coding sheets, source codes listings, functional
# specifications, specifications, user operation manuals and all other materials
# and documents necessary to enable a reasonably skilled programmer to maintain,
# amend and enhance the Program and whether in eye-readable or machine-readable
# form; and
#
# "IPRs" shall include all registered and unregistered patents, trademarks,
# service marks, database rights, design rights, copyright, applications for any
# of the foregoing, and any other similar protected rights in any country including
# the right to register or enforce the same.


# checking that SHELL is set
if [ -z "\$SHELL" ]
then
  # Using bash if SHELL is not one of the shells we know accept -c or if it is not executable, or if it is not installed
  _SN="`basename \$SHELL`"
  if [ \\( "\$_SN" !=  "bash" -a "\$_SN" != "csh" -a  "\$_SN" != "tcsh" -a "\$_SN" != "sh" -a "\$_SN" != "ksh" -a "\$_SN" != "zsh" \\) -o ! -x "$self->{config}{sw_root_path}/\$SHELL" ] 
  then
    SHELL="/bin/bash"
  fi
fi

if [ -z "\$*" ]
then
  exec "$qt_start" "\$SHELL"
else
  exec "$qt_start" "\$SHELL" "-c" "--" "\$*"
fi
};

  # Write out the script and link it to the bin
  my $runx86_path = "$self->{config}{qt_root_path}/bin/".$self->RUNX86;
  my $runx86_symlink = $self->BIN_DIR.'/'.$self->RUNX86;

  $self->write_file($runx86_path, $runx86_script, ':unix');

  chmod 0755, $runx86_path or
    $self->croak("Cannot chmod [_1]: [_2]", $runx86_path, $!);

  # Create the BIN_DIR if it doesn't already exist
  if (not -d $self->BIN_DIR) {
    File::Path::mkpath($self->BIN_DIR);
  }

  if (-l $runx86_symlink or -e $runx86_symlink) {
    unlink $runx86_symlink or
      $self->croak("Failed to remove [_1]", $runx86_symlink);
  }
  symlink($runx86_path, $runx86_symlink) or
    $self->croak("Failed to create symlink [_1]", $runx86_symlink);

  # Drop information about the QT in the SW file
  my $version_info = $self->run_command($qt_start, '-b', '2>&1');
  $self->write_file("$self->{config}{sw_root_path}/etc/".$self->SW_SUCC,
		    $version_info);

  # Tell the user we were successful
  $self->log;
  $self->log("To start a translated [_1] shell, run [_2]",
	     $self->SW_NAME, $self->BIN_DIR.'/'.$self->RUNX86);
  $self->log;

  # Return true
  return 1;
}

=item B<populate_sw_rpm_dir>

Create the RPM directory according to $self->{config}.
This is a temporary directory (defaults to /tmp/rpms)
in which we copy all RPMs before installation.

The RPMs are retrieved either from ISOs, or from CD/DVDs,
or from directories containing .rpm files.

Returns two values (as an array)
 1) The number of RPM files currently stored
 2) The number of RPM files that we require

So we're ready to go if (1) == (2)

=cut

sub populate_sw_rpm_dir {
  my ($self) = @_;

  #
  # See how many RPMs we have so far.
  #
  $self->analyze_sw_rpm_dir;

  while ($self->{config}{rpms_found} < $self->{config}{total_rpms_required}) {

    #
    # If the user enters no directory, or there's an error, then stop
    # looping for directory names.
    #
    my $one_or_more_files = $self->get_installation_media_config_interactively;
    last if ! $one_or_more_files or $one_or_more_files eq 'skip';

    #
    # IF the entered directory is 'list', then list the RPMs still required
    #
    if (lc $one_or_more_files eq 'list') {
      my @rpms = keys %{$self->{config}{rpms_still_to_get}};
      $self->log;
      $self->log("The following [_1] RPMs are still required:", scalar(@rpms));
      $self->log(" - [_1]", $_) for sort @rpms;
      next;
    }

    $self->{config}{rpm_files} = [];
    $self->{config}{iso_files} = [];
    foreach my $file_or_dir (glob($one_or_more_files)) {
      if (-f $file_or_dir) {
	if ($file_or_dir =~ /\.rpm$/) {
	  push @{$self->{config}{rpm_files}}, $file_or_dir;
	}
	elsif ($file_or_dir =~ /\.iso$/) {
	  push @{$self->{config}{iso_files}}, $file_or_dir;
	}
	else {
	  $self->log("The file [_1] is not a .iso or .rpm", $file_or_dir);
	}
      }
      elsif (-d $file_or_dir) {
	$self->log("Examining directory. Please wait...");
	$self->search_dir_for_rpms_and_isos($file_or_dir);
      }
      else {
	$self->log("Cannot find '[_1]'", [$file_or_dir, 'user']);
      }
    }

    #
    # Copy the RPM files we have
    #
    $self->copy_rpm_files_to_sw_rpm_dir;

    #
    # Find the RPMs in .iso files
    #
    foreach my $iso (@{$self->{config}{iso_files}}) {
      if (not -d $self->ISO_MOUNTPOINT) {
	File::Path::mkpath($self->ISO_MOUNTPOINT);
      }
      my $success = $self->mount_iso($iso, $self->ISO_MOUNTPOINT);
      last if not defined $success; # undef means we can't mount at all
      next if not $success; # 0 means we can't mount this iso
      $self->log("Examining ISO file [_1]", [$iso, 'user']);
      $self->search_dir_for_rpms_and_isos($self->ISO_MOUNTPOINT,
					  no_isos => 1);
      my $num_rpms_found = @{$self->{config}{rpm_files}};
      $self->copy_rpm_files_to_sw_rpm_dir;
      $self->umount($self->ISO_MOUNTPOINT);

      #
      # Don't waste time searching more ISOs unless we're done
      #
      $self->analyze_sw_rpm_dir;
      last if $self->{config}{rpms_found} == $self->{config}{total_rpms_required};
    }
    $self->{config}{iso_files} = [] ; # so we don't re-search

    $self->analyze_sw_rpm_dir;
    $self->log("Got [_1]/[_2] RPMs", $self->{config}{rpms_found},
	       $self->{config}{total_rpms_required});
    $self->log;

  }

  return ($self->{config}{rpms_found} ==
	  $self->{config}{total_rpms_required} ? 1 : 0);
}

=item B<get_installation_media_config_interactively>

Asks the user for a directory containing the RPMs.

Sets $self->{config}{sw_rpm_source_dir}

=cut

sub get_installation_media_config_interactively {
  my ($self) = @_;
  $self->log(<<END, $SUBJ_WORLDS{$self->{config}{sw_distro}}{title});
Please enter the directory or files containing [_1].
 - ISO9660 images are accepted (if they end with .iso)
 - RPM files are accepted (if they end with .rpm)
 - If you have a CD/DVD-ROM, please mount it and enter the mount path

You can enter each file or mount path individually or provide a pattern to
select multiple files at once, e.g. /files/rhel/RHEL4-U4-i386-AS-disc*.iso.

To view a list of missing RPM files, type 'list'
To return to the main menu, type 'skip'

END

  return $self->ask('q'  => 'Enter directory or file:',
		    type => 'path', list => 1, skip => 1);
}

=item B<get_rpm_list_dir>

Get the directory containing the RPM lists for the chosen subject
world.

=cut

sub get_rpm_list_dir {
  my ($self) = @_;
  return $self->ROOT.
    "/resources/sw_lists/$self->{product}/$self->{config}{sw_distro}";
}

=item B<get_rpm_md5sum_file>

Get the file containing MD5 checksums of each RPM.

=cut

sub get_rpm_md5sum_file {
  my ($self) = @_;
  return $self->ROOT.
    "/resources/sw_lists/$self->{product}/$self->{config}{sw_distro}.md5sum";
}

=item B<get_rpm_list_as_hashref>

Get a list of all the RPMs we need to copy across for the selected
SW. Return it as a hashref with the RPMs as keys for easy
lookup/removal.

Supply a filename or glob as a second parameter if * is too greedy.

=cut

sub get_rpm_list_as_hashref {
  my ($self, $filename) = @_;
  my $rpmlist_dir = $self->get_rpm_list_dir;
  if (not -d $rpmlist_dir) {
    $self->croak("Cannot find RPM list directory [_1]", $rpmlist_dir);
  }
  $filename ||= '*';
  my %rpms_hash;
  foreach my $file (glob("$rpmlist_dir/$filename")) {
    next unless -f $file; # skip directories (eg. CVS/)
    my @rpms = split /\s+/, $self->slurp($file);

    # There will be some extra info in this file that won't correspond to
    # an rpm file, so filter it out now.
    @rpms = grep !/^--/, @rpms;

    # slice the hash with @rpms creating all the keys with undefs as the values
    @rpms_hash{@rpms} = ();
  }
  if (not %rpms_hash) {
    $self->croak("Cannot find any RPM lists in [_1]", $rpmlist_dir);
  }

  return \%rpms_hash;
}

=item B<search_dir_for_rpms_and_isos>

Do a recursive find on the provided directory for any .rpm
files and .iso files.

Optional arguments (provided as hash):
 - no_isos - don't search for ISOs. Used when we're searching
   inside an ISO, as we don't have the ability to recursively
   mount ISOs.

=cut

sub search_dir_for_rpms_and_isos {
  my ($self, $dir, %args) = @_;
  # Follow a symlink if it's the very top level specified. Don't
  # follow symlinks below this (using find's 'follow' option) as it
  # can be dangerous.
  if (-l $dir) {
    return $self->search_dir_for_rpms_and_isos(readlink $dir, %args);
  }
  my $find_sub = sub {
    $File::Find::prune = 1 if (-d and (not -x or not opendir(my $d, $_)));
    return unless -f and -r;
    if (/\.rpm$/) {
      my $basename = File::Basename::basename($File::Find::name);
      if (exists $self->{config}{rpms_still_to_get}{$basename}
	 or $self->check_regex_rpms_for_file($basename)) {
	push @{$self->{config}{rpm_files}}, $File::Find::name;
      }
    }
    elsif (/\.iso$/ and not $args{no_isos}) {
      push @{$self->{config}{iso_files}}, $File::Find::name;
    }
  };
  File::Find::find({wanted => $find_sub, no_chdir => 1}, $dir);
}

=item B<check_regex_rpms_for_file>

Check the regex_rpms list to see if the given file basename matches
one of the regexes. If it does, the following happens:

  1. The key in rpms_still_to_get matching the regex is changed to the
  real filename.

  2. The real filename is added to the real_names_for_regexes hash
  with the regex as the key.

  3. The regex is removed from the regex_rpms list.

  4. True is returned.

Otherwise, false is returned.

=cut

sub check_regex_rpms_for_file {
  my ($self, $file) = @_;
  foreach my $regex (@{$self->{config}{regex_rpms}}) {
    if (exists $self->{config}{rpms_still_to_get}{"/$regex/"}
	and $file =~ /^$regex$/) {
      # Replace with its real name
      delete $self->{config}{rpms_still_to_get}{"/$regex/"};
      $self->{config}{rpms_still_to_get}{$file} = undef;
      # Record the name to use when we encounter the regex again
      $self->{config}{real_names_for_regexes}{$regex} = $file;
      # Remove it from the list
      @{$self->{config}{regex_rpms}} =
	grep {$_ ne $regex} @{$self->{config}{regex_rpms}};
      return 1;
    }
  }
  return 0;
}

=item B<analyze_sw_rpm_dir>

Looks in $self->{config}{sw_rpm_dir} to see all the RPMs we have.
Creates a list of those missing, and puts it in
$self->{config}{rpms_still_to_get}.

Sets: $self->{config}{rpms_still_to_get} (array or strings)
      $self->{config}{rpms_found}        (integer, count)

=cut

sub analyze_sw_rpm_dir {
  my ($self) = @_;

  opendir DIR, $self->{config}{sw_rpm_dir}
    or $self->croak('Cannot find the directory [_1]',
		    $self->{config}{sw_rpm_dir});
  foreach (readdir DIR) {
    my $file_and_dir = "$self->{config}{sw_rpm_dir}/$_";
    next unless -f $file_and_dir;
    if (exists $self->{config}{rpms_still_to_get}{$_}
	or $self->check_regex_rpms_for_file($_)) {
      delete $self->{config}{rpms_still_to_get}{$_};
      $self->{config}{rpms_found}++;
    }
  }

  closedir DIR or $self->croak("Cannot close directory connection");
}

=item B<copy_rpm_files_to_sw_rpm_dir>

Copies all files in $self->{config}{rpm_files}
into $self->{config}{sw_rpm_dir}

=cut

sub copy_rpm_files_to_sw_rpm_dir {
  my ($self) = @_;
  foreach (@{$self->{config}{rpm_files}}) {
    my $basename = File::Basename::basename $_;
    if (exists $self->{config}{rpms_still_to_get}{$basename}) {

      $self->check_rpm_md5sum($_, $basename) or do {
        $self->log("MD5 checksum of '[_1]' is incorrect - it is probably corrupt.", $basename);
	next;
      };

      File::Copy::copy($_, $self->{config}{sw_rpm_dir})
	  or $self->croak("Cannot copy [_1] to [_2]", $_,
			  $self->{config}{sw_rpm_dir});
    }
  }

  $self->{config}{rpm_files} = []; # so we don't re-copy
}

=item B<check_rpm_md5sum>

  my $ok = $self->check_rpm_md5sum($full_file_name, $basename);

Given an RPM file, returns TRUE iff either the md5sum matches
that stored in the md5sum file, or the file isn't present in the md5sum file.

=cut

sub check_rpm_md5sum {
  my ($self, $full_name, $basename) = @_;

  #
  # Read in the md5sums, unless we've done it before.
  #
  unless (exists $self->{rpm_md5sums}) {
    $self->{rpm_md5sums} = {};

    my $md5sum_file = $self->get_rpm_md5sum_file;
    if (!-f $md5sum_file) {
      $self->log_only_to_file("Cannot find '$md5sum_file'. Carrying on without it.\n");
      return 1; # if no md5sums, carry on anyway
    }

    my @contents = $self->slurp($md5sum_file);
    foreach (@contents) {
      (/^([\w\d]{32})\s+(.+?\.rpm)\s*$/)
        or $self->croak("Could not understand line in '[_1]': [_2]", $md5sum_file, $_);
      $self->{rpm_md5sums}{$2} = $1;
    }
  }

  #
  # No md5sum for this file? We assume everything's OK
  #
  my $stored_md5sum = $self->{rpm_md5sums}{$basename};
  if (not $stored_md5sum) {
    $self->log_only_to_file("No md5sum to check $basename against. Assuming it's OK.\n");
    return 1;
  }

  #
  # Work out the md5sum of this file and compare against the stored one
  #
  my $this_md5sum = $self->md5sum_of_file($full_name);
  return 1 if ($this_md5sum eq $stored_md5sum);

  $self->log_only_to_file("Checksums differ for $basename (expected '$stored_md5sum', got '$this_md5sum')\n");
  return 0;
}

=item B<check_disk_space_in_sw_rpm_dir>

Returns true if there is enough disk space in the sw_rpm_dir for the
RPMs.

If not, alerts the user and returns false.

=cut

sub check_disk_space_in_sw_rpm_dir {
  my ($self) = @_;
  my $free = $self->check_disk_space_in($self->{config}{sw_rpm_dir});
  my $required = ($SUBJ_WORLDS{$self->{config}{sw_distro}}{rpmsize} * 1.1); # add 10% for safety

  if ($free < ($required*1024)) {
    $self->log("Disk space is too low to create temporary directory [_1].",
	       $self->{config}{sw_rpm_dir});
    $self->log("Please make sure there are at least [_1] MB free before ".
	       "retrying.", $required);
    return 0;
  }

  return 1;
}

=item B<check_disk_space_in_sw_root_dir>

Returns true if there is enough disk space in the sw_root_path for the
RPMs.

If it is on the same disk as the RPM dir, then it will expect the
combined size to be free.

If not, alerts the user and returns false.

=cut

sub check_disk_space_in_sw_root_dir {
  my ($self) = @_;
  my $free = $self->check_disk_space_in($self->{config}{sw_root_path});
  my $required = ($SUBJ_WORLDS{$self->{config}{sw_distro}}{size} * 1.1); # add 10% for safety

  # Add on the rpmsize if the RPMs are to be stored on the same
  # filesystem.
  if ($self->are_files_on_same_filesystem($self->{config}{sw_rpm_dir},
					  $self->{config}{sw_root_path})) {
    $required += $SUBJ_WORLDS{$self->{config}{sw_distro}}{rpmsize};
  }

  if ($free < ($required*1024)) {
    $self->log("Disk space is too low to create [_1] in [_2].",
	       $self->SW_NAME, $self->{config}{sw_root_path});
    $self->log("Please make sure there are at least [_1] MB free before ".
	       "retrying.", $required);
    return 0;
  }

  return 1;
}



=item B<write_config_file>

Write out a config file with any options that are different to the defaults.

=cut

sub write_config_file {
  my ($self) = @_;

  # Common
  my @config = ($self->get_uncommon_config_lines,
		$self->get_config_file_lines_common);

  # Other things
  if ($self->{config}{qt_root_path} ne $self->QT_ROOT_PATH) {
    push @config,
      "LOCALISATION_FILES_DIR=$self->{config}{qt_root_path}/locale";
  }

  if (@config) {
    $self->write_file($self->CONFIG_FILE, \@config, ':unix');
  }
  elsif (-f $self->CONFIG_FILE) {
    unlink $self->CONFIG_FILE or
      $self->croak("Cannot remove [_1]: [_2]", $self->CONFIG_FILE, $!);
  }
}

=item B<get_upgrade_config_interactively>

Check if we already have QT installed and then ask the user which
version to upgrade to.

=cut

sub get_upgrade_config_interactively {
  my ($self) = @_;

  my $action = $self->prompt_user_for_upgrade;
  return 0 if $action eq 'return';

  # QT upgrade
  if ($action eq 'qt') {
    $self->{config}{upgrade_qt} = 1;

    # Check QT is already installed
    if (not $self->{config}{rpm_installed}) {
      $self->log("Cannot upgrade [_1] because it is not installed.",
 		 $self->PRODUCT_NAME);
      return 0;
    }

    # Ask the user which version should be installed
    $self->log("The currently installed version is: [_1]",
               $self->{config}{rpm_installed});
    $self->{config}{chosen_rpm_file} = $self->version_menu;
    return 0 if $self->{config}{chosen_rpm_file} eq 'return';
  }

  return 1;
}

=item B<prompt_user_for_upgrade>

Submenu for 'upgrade'.

=cut

sub prompt_user_for_upgrade {
  my ($self) = @_;
  my $options = [
		 [qt     => ["Upgrade a [_1] installation.",
			     $self->PRODUCT_NAME]],
		 [return => "Go back to the Main Menu."],
		 [quit   => "Quit."],
		];
  my $action = $self->menu($options);
  $self->quit if $action eq 'quit';
  return $action;
}

=item B<upgrade>

Install the new version of QT if appropriate.

=cut

sub upgrade {
  my ($self) = @_;

  $self->stop_qt;

  # Reinstall the installer
  my $installer_path = "$self->{config}{qt_root_path}/installer";
  if (-d $installer_path) {
    File::Path::rmtree($installer_path);
  }
  $self->install_self_into($installer_path);

  # Re-symlink the doc/ directory in case it's moved
  my $doc_path = "$self->{config}{qt_root_path}/doc";
  if (-l $doc_path) {
    unlink $doc_path;
  }
  symlink "$installer_path/resources/doc/RO", $doc_path;

  # RPM
  $self->install_rpm($self->{config}{chosen_rpm_file},
		     upgrade => 1,
		     relocate => [$self->QT_ROOT_PATH,
				  $self->{config}{qt_root_path}]) or
    $self->croak("RPM upgrade failed.");

  $self->start_qt;
  $self->log;
  $self->log("Upgrade successful.");
  $self->log;

  if ($self->{config}{rpm_installed} and $self->{config}{sw_installed}) {
    $self->bind_qt_and_sw or return 0;
  }

  # Success! Add to transactions
  $self->add_transaction("Upgraded [_1]", $self->PRODUCT_NAME);
}

=item B<get_uninstall_config_interactively>

Ask the user what they want to uninstall and make sure it's safe to do
so.

=cut

sub get_uninstall_config_interactively {
  my ($self) = @_;
  my $action = $self->prompt_user_for_uninstall;
  return 0 if $action eq 'return';

  # Remember what the user wants to uninstall
  $self->{config}{uninstall_qt} = ($action =~ /qt/);
  $self->{config}{uninstall_sw} = ($action =~ /sw/);

  # Check things about QT
  if ($self->{config}{uninstall_qt}) {
    if (not $self->{config}{rpm_installed}) {
      $self->log("Cannot uninstall [_1] because it is not installed.",
		$self->PRODUCT_NAME);
      return 0;
    }
    $self->check_qt_is_not_running or return 0;
    # Get confirmation
    $self->ask(q => ["Are you sure you want to uninstall [_1]?",
		     $self->PRODUCT_NAME],
	       type => 'bool', default => 1) or return 0;

    # Delete the logfile?
    if ( $self->ask(q => ["Remove log file directory ([_1])?",
		     [$self->{config}{logfile_path}, 'user']], type => 'bool', default => 0) ) {
		  $self->{config}{remove_logfile} = 1 ;
		}
	       	       
  }

  # Check things about SW
  if ($self->{config}{uninstall_sw}) {
    if (not -d $self->{config}{sw_root_path}) {
      $self->log("Cannot uninstall [_1] because it is not installed.",
		$self->SW_NAME);
      return 0;
    }
    $self->check_qt_is_not_running or return 0;
    if ($self->{config}{sw_installed}) {
      $self->ask(q => ["Delete [_1] (installed at [_2]) from the filesystem?\nYou will lose all custom config and installed applications.",
		       $self->SW_NAME, [$self->{config}{sw_root_path}, 'user']],
		 type => 'bool', default => 0) or return 0;
    }
    elsif (-f "$self->{config}{sw_root_path}/etc/".$self->SW_PART) {
      $self->ask(q => ["Delete partially installed [_1] at [_2] from the filesystem?",
		       $self->SW_NAME, [$self->{config}{sw_root_path}, 'user']],
		 type => 'bool', default => 0) or return 0;
    }
    else {
	$self->ask(q => ["Directory [_1] exists but doesn't look like an installation of [_2]. Delete it anyway?",
			[$self->{config}{sw_root_path}, 'user'], $self->SW_NAME],
		   type => 'bool', default => 0) or return 0;
    }
    my $mounted = $self->directory_has_mounted_filesystems
      ($self->{config}{sw_root_path});
    if ($mounted) {
      $self->log("The directory [_1] contains mounted filesystems ([_2]). Please unmount them before retrying this operation.", [$self->{config}{sw_root_path}, 'user'], $mounted);
      return 0;
    }
  }

  # Will either component be left at the end of this installation?
  # If not, ask if installation config dir will be removed
  if (($self->{config}{uninstall_qt} || !$self->{config}{rpm_installed}) &&
      ($self->{config}{uninstall_sw} || !$self->{config}{sw_installed}) ) {
    $self->{config}{uninstalled_everything} = 1;
    if ( $self->ask(q => ["Remove the installation config directory ([_1])?",
         $self->CONFIG_DIR], type => 'bool', default => 0) ) {
	    $self->{config}{remove_config_dir} = 1 ;
    }
  }

  return 1;
}

=item B<prompt_user_for_uninstall>

Submenu for 'uninstall'.

=cut

sub prompt_user_for_uninstall {
  my ($self) = @_;
  my $options = [
		 [qt_sw => ["Uninstall [_1] + [_2].",
	     $self->PRODUCT_NAME, $self->SW_NAME]],
		 [qt => ["Uninstall [_1].",
			 $self->PRODUCT_NAME]],
		 [sw => ["Uninstall [_1].",
			 $self->SW_NAME]],
		 [return => "Go back to the Main Menu."],
		 [quit   => "Quit."],
		];
  my $action = $self->menu($options);
  $self->quit if $action eq 'quit';
  return $action;
}

=item B<uninstall>

Uninstall QT and/or SW as appropriate.

=cut

sub uninstall {
  my ($self) = @_;
  if ($self->{config}{uninstall_qt}) {
    $self->log("Uninstalling [_1]. Please wait...", $self->PRODUCT_NAME);
    $self->stop_qt;
    $self->uninstall_rpm($self->PRODUCT_NAME) or
      $self->croak("Failed to uninstall [_1] RPM", $self->PRODUCT_NAME);

    #Remove the qt directory
    if ( -e $self->{config}{qt_root_path} ) {
      File::Path::rmtree($self->{config}{qt_root_path});
      $self->log("[_1] installation directory ([_2]) deleted.", 
      $self->PRODUCT_NAME, [$self->{config}{qt_root_path}, 'user']);
    }

    #Remove the log directory
    if ($self->{config}{remove_logfile}) {
      File::Path::rmtree($self->{config}{logfile_path});
      $self->log("Log file directory ([_1]) deleted.", [$self->{config}{logfile_path}, 'user']);
    }
    else {
      $self->log("Please remove log file directory ([_1]) if you no longer require it",
                  [$self->{config}{logfile_path}, 'user']);
    }
    # devLog created by QT
    unlink "/var/opt/p-ave/devLog";
    $self->add_transaction("Uninstalled [_1]", $self->PRODUCT_NAME);
  }
  if ($self->{config}{uninstall_sw}) {
    # This is a scary thought, but we have asked the user if he is
    # sure by this stage...
    $self->log("Uninstalling [_1]. Please wait...", $self->SW_NAME);
    File::Path::rmtree($self->{config}{sw_root_path});
    if (-e $self->BIN_DIR.'/'.$self->LINKX86) {
      unlink $self->BIN_DIR.'/'.$self->LINKX86 or
	$self->croak("Can't remove [_1]: [_2]", $self->LINKX86, $!);
    }
    $self->add_transaction("Uninstalled [_1]", $self->SW_NAME);
  }
  if (-e $self->BIN_DIR.'/'.$self->RUNX86) {
    unlink $self->BIN_DIR.'/'.$self->RUNX86 or
	$self->croak("Can't remove [_1]: [_2]", $self->RUNX86, $!);
  }

  if ($self->{config}{uninstalled_everything}) {
    # Remove the installer configuration directory
    if ($self->{config}{remove_config_dir}) {
      File::Path::rmtree($self->CONFIG_DIR);
      $self->log("Config directory ([_1]) deleted.", $self->CONFIG_DIR);
    }
    else {
      $self->log("Please remove config directory ([_1]) if you no longer require it", $self->CONFIG_DIR);
    }
    $self->quit;
  }


}

=item B<check_qt_is_not_running>

Returns true if QT (PRODUCT_NAME) is not running, ignoring zombie processes.

If it is running, will print out a warning about uninstallation being
impossible and return false.

=cut

sub check_qt_is_not_running {
  my ($self) = @_;
  my @qt_pids = $self->get_running_processes_of_type($self->PRODUCT_NAME);
  if (@qt_pids) {
    my @pids = $self->get_living_processes(@qt_pids);
    if (@pids) {
      $self->log;
      $self->log("The following translated processes (pids) are running:");
      $self->log("@pids"); # quotes are necessary here
          $self->log;
      $self->log("You cannot uninstall while there are still translated processes running.\nKill these processes to continue.");
      return 0;
    }
  }
  return 1;
}

=item B<start_qt>

Start the translator daemon. Croaks if there was a problem.

=cut

sub start_qt {
  my ($self) = @_;
  my @command = ('/etc/init.d/'.$self->PRODUCT_NAME, 'start');
  my $output = $self->run_command(@command, '2>&1');
  if ($self->{exit}) {
    # Only log the output to the screen if it fails
    $self->log($output);
    $self->croak("Problem starting [_1]. Exit value was [_2].",
		 $self->PRODUCT_NAME, $self->{exit});
  } else {
    $self->log_only_to_file($output);
  }
}

=item B<stop_qt>

Stop the translator daemon. Croaks if there was a problem.

=cut

sub stop_qt {
  my ($self) = @_;
  # We might be calling this even if we haven't got one to stop
  return if not -x '/etc/init.d/'.$self->PRODUCT_NAME;

  my @command = ('/etc/init.d/'.$self->PRODUCT_NAME, 'stop');
  my $output = $self->run_command(@command, '2>&1');
  if ($self->{exit}) {
    $self->log($output);
    $self->croak("Problem stopping [_1]. Exit value was [_2].",
		 $self->PRODUCT_NAME, $self->{exit});
  } else {
    $self->log_only_to_file($output);
  }
}

=item B<restart_qt>

Restart the translator daemon. Croaks if there was a problem.

=cut

sub restart_qt {
  my ($self) = @_;
  my @command = ('/etc/init.d/'.$self->PRODUCT_NAME, 'restart');
  my $output = $self->run_command(@command, '2>&1');
  if ($self->{exit}) {
    $self->log($output);
    $self->croak("Problem restarting [_1]. Exit value was [_2].",
		 $self->PRODUCT_NAME, $self->{exit});
  } else {
    $self->log_only_to_file($output);
  }
}

=item B<run_translated>

Like run_command() but will invoke the given command with the
translator.

Must be an absolute *subject world* path, eg. /bin/ls

Your current working directory must be visible in the subject world or
the translator will croak.

=cut

sub run_translated {
  my ($self, $argv0, @argv) = @_;
  my @qt = ("$self->{config}{qt_root_path}/bin/".$self->PRODUCT_NAME);
  $argv0 = $self->{config}{sw_root_path}.$argv0;
  $self->log_only_to_file(POSIX::strftime('%H:%M:%S', localtime).
			  " [TRANSLATED $argv0 @argv]\n",
			 undef, $self->{encoding});
  if (defined wantarray) {
    return $self->run_command(@qt, $argv0, @argv);
  }
  else {
    $self->run_command(@qt, $argv0, @argv);
  }
}

=item B<get_installed_products>

Check if QT or SW are installed and return them as a list.

=cut

sub get_installed_products {
  my ($self) = @_;
  my @products;
  if ($self->{config}{rpm_installed}) {
    push @products, ["[_1] from package [_2]", $self->PRODUCT_NAME,
		     $self->{config}{rpm_installed}];
  }
  if ($self->{config}{sw_installed}) {
    push @products, ["[_1] ([_2]) installed at [_3]", $self->SW_NAME,
		     $self->{config}{sw_installed},
		     [$self->{config}{sw_root_path}, 'user']];
  }
  return @products;
}

=item B<get_configure_config_interactively>

Want to bind NIS?

=cut

sub get_configure_config_interactively {
  my ($self) = @_;
  my %config;
  if ($self->{config}{sw_installed}) {
    my $answer = $self->menu([
		 [bind_nis => ["Bind NIS for [_1].", $self->SW_NAME]],
		 [return   => "Go back to the Main Menu."]
		]);
    if ($answer ne 'return') {
      $config{$answer} = 1;
    }
    return \%config;
  }	
  else {
    return $self->SUPER::get_configure_config_interactively;
  }
}

=item B<configure>

Bind NIS if appropriate.

=cut

sub configure {
  my ($self, $config) = @_;
  if ($config->{bind_nis}) {
    $self->bind_nis;
  }
}

=item B<bind_nis>

Bind NIS from host machine into SW.

=cut

sub bind_nis {
  my ($self) = @_;
  $self->log("Binding NIS...");

  # Check how the host machine is configured first, and generate a
  # %subst with substitutions
  my %subst;
  my @nsswitch = $self->slurp('/etc/nsswitch.conf');
  s/#.*$// for @nsswitch; # Throw away comments

  foreach my $line (@nsswitch) {
    my ($key, $value) = $line =~ /^\s*(\S+)\s*:\s*(.+?)\s*$/;
    next unless $key and $value;

    # If nis is enabled in compat mode, we need to change it
    if ($key =~ /^(passwd|group|shadow)$/ and $value =~ /compat/) {
      my @file = $self->slurp("/etc/$key");
      # If we see a ^+ line in the file then nis has been enabled in
      # compat mode. Switch to 'nis files' in SW.
      if (grep /^\+/, @file) {
	$subst{$key} = 'nis files';
      }
      # If not, then compat mode is on without nis, so switch to
      # 'files' in SW.
      else {
	$subst{$key} = 'files';
      }
    }

    # For everything else we just need to check if 'nis' appears
    # before 'files' or the other way around.
    my $nis_idx = index $line, 'nis';
    my $files_idx = index $line, 'files';
    if ($nis_idx >= 0 and $files_idx >= 0) {
      $subst{$key} = ($nis_idx < $files_idx ? 'nis files' : 'files nis');
    }
  }

  # Now we open the SW nsswitch and make the changes
  my @sw_nsswitch =
    $self->slurp("$self->{config}{sw_root_path}/etc/nsswitch.conf");
  foreach (@sw_nsswitch) {
    next if /nis/; # don't touch lines that already have nis
    my ($key) = /^\s*(\S+)\s*:/;
    if ($key and $subst{$key}) {
      s/(files|compat)/$subst{$key}/; # substitute the first files/compat
    }
  }
  # Write it back
  $self->write_file("$self->{config}{sw_root_path}/etc/nsswitch.conf",
		    \@sw_nsswitch);

  # Also escape the /var/yp/binding directory if not done already
  $self->escape("/var/yp/binding");
  $self->log("Done.");
  $self->add_transaction("Bound NIS into [_1]", $self->SW_NAME);
}

=item B<versions_available>

Returns an hashref of the available versions to install on this
system.

The keys are version names (eg. 0.3.2-SLES) and the values are the
path of the RPM to install.

Croaks if no RPM can be found.

=cut

sub versions_available {
  my ($self) = @_;
  my $dir = $self->ROOT.'/resources/rpm/RO';
  if (not -d $dir) {
    $self->croak("No such directory [_1].", $dir);
  }
  my $rpm_glob = $self->PRODUCT_NAME.
    "-*-$self->{distro_details}{rpm}.ppc64.rpm";
  my $version_regex = '([^/]*).ppc64.rpm$';
  # We need to chdir here in case the directory name contains spaces
  my $cwd = Cwd::cwd;
  chdir $dir or $self->croak("Couldn't chdir");
  my @files = glob($rpm_glob);
  chdir $cwd or $self->croak("Couldn't chdir");
  if (not @files) {
    $self->croak("No RPM file for [_1] could be found",
		 $self->{distro_details}{rpm});
  }
  my %versions;
  foreach (@files) {
    my $dir_and_file = "$dir/$_";
    my $version = $self->get_uninstalled_rpm_full_name($dir_and_file);
    if (defined $version) {
      $versions{$version} = $dir_and_file;
    }
    else {
      $self->croak("RPM is unable to query [_1]", $dir_and_file);
    }
  }

  # Delete the current version
  if ($self->{config}{rpm_installed}) {
    delete $versions{$self->{config}{rpm_installed}};
  }

  return \%versions;
}

=item B<version_menu>

Give the user a choice of versions from
$self->{config}{versions_available}.

=cut

sub version_menu {
  my ($self) = @_;
  my $options = [];
  foreach my $version (sort keys %{$self->{config}{versions_available}}) {
    push @$options,
      [$self->{config}{versions_available}{$version} => $version];
  }
  if (not @$options) {
    $self->log;
    $self->log("Nothing is available to be installed on your system.");
    return 'return';
  }
  push @$options, [return => "Go back to the Main Menu."];
  return $self->menu
    (
     $options,
     ["Please select the version of [_1] you would like to install:",
      $self->PRODUCT_NAME]
    );
}

=item B<welcome>

Additional info for the welcome message.

=cut

sub welcome {
  my ($self) = @_;
  $self->SUPER::welcome;
  my $welcome = q{
To run Linux/x86 applications the installation will require Linux/x86
application binaries, libraries and infrastructure files as well as the
PowerVM Lx86 (System p AVE) software.

Please refer to the PowerVM Lx86 (System p AVE) Administration Guide and
Release Notes for further details of the installation requirements.

};
  $self->log($welcome);
}

=item B<ibm_registration_text>

Returns a string containing the contents of the registration
email to be sent to IBM.

=cut

sub ibm_registration_text {
  my ($self) = @_;
  my $c = $self->{config};
  my $text = qq{
Date:         $c->{registration_date}
System:       $c->{model}
Serial:       $c->{serial}

Company:      $c->{ibm_company_name}

Name:         $c->{ibm_registrant_name}
Address:      $c->{ibm_address}
Phone Number: $c->{telephone_number}
Email:        $c->{ibm_email_address}
};

  return $text;
}

=item B<ibm_subject_line>

Returns a string containing the subject line for the 
registration email to be sent to IBM.

=cut

sub ibm_subject_line {
  my ($self) = @_;

  my $subject_text = 'Activation of PowerVM Lx86 (System p AVE) from: ' . 
                       $self->{config}->{ibm_company_name} ;
  
  return $subject_text ;  
}

=item B<escape>

Given a real absolute path, escape it into the subject world.

If it doesn't exist on the target system, skip it silently.

If it exists in the subject world, remove it first.

Given an array, does this for each argument.

=cut

sub escape {
  my ($self, @path) = @_;
  if (not $self->{config}{sw_root_path}) {
    $self->croak("sw_root_path is not set");
  }
  foreach my $path (@path) {
    next unless -e $path;
    if (not $path or $path !~ /\w/) {
      $self->croak("path '[_1]' is invalid", $path);
    }
    my $swpath = "$self->{config}{sw_root_path}$path";

    # If it already exists, remove it and any subdirs
    if (-e $swpath) {
      File::Path::rmtree($swpath);
    }
    # Otherwise create the path to it but remove it itself.
    else {
      File::Path::mkpath($swpath);
      rmdir $swpath;
    }

    # Symlink it in
    symlink $path, $swpath or
      $self->croak("Couldn't symlink [_1] to [_2]: [_3]", $path, $swpath, $!);
  }
}

=item B<get_registration_info_interactively>

This function collects common registration info interactively and adds
it to $self->{config}.

=cut

sub get_registration_info_interactively {
  my ($self) = @_;
  my $success1 = $self->get_ibm_registration_info_interactively;
  my $success2 = $self->get_distro_registration_info_interactively;
  return $success1 && $success2;
}

=item B<get_ibm_registration_info_interactively>

Called exclusively by get_registration_info_interactively.
Sends the email to IBM.

=cut

sub get_ibm_registration_info_interactively {
  my ($self) = @_;
  my @qs =
    (
     [ibm_company_name => "Please enter the name of your company:",
     '.+', 'A company name must be provided.'],
     [ibm_registrant_name => "Please enter your name (optional):"],
     [ibm_address => "Please enter your address, use commas to separate fields (optional):"],
     [telephone_number => "Please enter your phone number (optional):"],
     [ibm_email_address => "Please enter your email address (optional):",
      $self->EMAIL_REGEXP . '?',
      'Email addresses must not contain spaces, must contain an @ and have a valid domain.'],
    );

  {
    # POSIX::strftime isn't sensitive to 'no locale' (never mind what
    # man perllocale says) so we need to set the LC_TIME locale back
    # to 'C' temporarily here so the string is in English.
    my $old = POSIX::setlocale(&POSIX::LC_TIME, 'C');
    $self->{config}{registration_date} =
      POSIX::strftime(DATE_FORMAT, localtime);
    POSIX::setlocale(&POSIX::LC_TIME, $old);
  }

  my $initial_text = qq{
You must provide some information to IBM before starting the installation
process.  This Activation information will be sent to IBM as an email.

Note: You must provide your company name for the installation to proceed.
All other details are optional.  The Activation email will also include
some details about this system.  Once you have entered this information
it will be presented to you to review.  You will have the option to
 send the information to IBM via email or abort the installation.

};

  return $self->register_interactively
    (
     email                       => $self->IBM_REGISTER_EMAIL,
     organisation                => 'IBM',
     record_file                 => $self->IBM_REG_FILE,
     questions                   => \@qs,
     subject_line_method         => 'ibm_subject_line',
     email_body_method           => 'ibm_registration_text',     
     initial_text_for_first_time => $initial_text,
     compulsory                  => 1, 
    );
}

=item B<get_distro_registration_info_interactively>

Called exclusively by get_registration_info_interactively.
Override if you need it.

=cut

sub get_distro_registration_info_interactively {
  my ($self) = @_;
  # By default, do nothing
}

=item B<goodbye>

Override this if your product must say something specific before exiting.

=cut

sub goodbye {
  my ($self) = @_;
  if ($self->{successful_registrations}{IBM}) {
    $self->log(qq{
The Activation email data is stored on the system in
  [_1].

If this system is not set up to send email, please email
this data file to IBM at <[_2]>.
}, $self->IBM_REG_FILE, $self->IBM_REGISTER_EMAIL);
  }

  if ($self->{successful_registrations}{'Red Hat Inc.'}) {
    $self->log(qq{
The Red Hat Subscription email data is stored on the system in
  [_1].

If this system is not set up to send email or you do not get an
acknowledgement email from Red Hat within the next 24 hours, please
email this data file to Red Hat at <[_2]>.
This will ensure that your Red Hat subscription is upgraded.
}, $self->RH_REG_FILE, $self->RH_REGISTER_EMAIL);
  }

  $self->SUPER::goodbye;

  # Finally
  $self->copy_log_file_to_installed_log_dir_if_appropriate;
}

=item B<copy_log_file_to_installed_log_dir_if_appropriate>

If the installed log dir exists at this stage, copy the installer
logfile there.

=cut

sub copy_log_file_to_installed_log_dir_if_appropriate {
  my ($self) = @_;
  if ($self->{config}{logfile_path} and -d $self->{config}{logfile_path}) {
    my $basename = File::Basename::basename($self->{logfile}->filename);
    $self->copy_preserve($self->{logfile}->filename,
			 "$self->{config}{logfile_path}/$basename");
  }
}

=item B<modify_login_defs>

Changes UID_MIN and GID_MIN in /etc/login.defs in the subject world to 50000.

=cut

sub modify_login_defs {
  my $self = shift;
  my $login_defs_file = "$self->{config}{sw_root_path}/etc/login.defs";
  my @login_defs = $self->slurp($login_defs_file);
  for my $line (@login_defs) {
    if ($line =~ /^((U|G)ID_MIN\s+)\d+/) {
      $line = "${1}50000";
    }
  }
  $self->write_file($login_defs_file, \@login_defs);
}

=item B<enable_services>

Enable translated services

=cut

sub enable_services {
  my ($self) = @_;
  # lets enable the basic required services now
  my $scripts_dir = $self->{config}{sw_root_path}.$self->INIT_D;
  chdir $self->{config}{sw_root_path} or $self->croak("Couldn't chdir");
  if( -x "$scripts_dir/dbus" ) {
    $self->run_translated('/sbin/chkconfig', 'dbus', 'on');
  }
  if( -x "$scripts_dir/syslog" ) {
    $self->run_translated('/sbin/chkconfig', 'syslog', 'on');
  }
  if( -x "$scripts_dir/crond" ) {
    $self->run_translated('/sbin/chkconfig', 'crond', 'on');
  } elsif( -x "$scripts_dir/cron" ) {
    $self->run_translated('/sbin/chkconfig', 'cron', 'on');
  }
}

=item B<check_login_defs>

  $self->check_login_defs($rpm);

Checks whether $rpm is known to modify /etc/login.defs in the subject world
and, if so, calls modify_login_defs to restore UID_MIN and GID_MIN to 50000.

Requires an override for each OS, because the rpm will be different.

=cut

sub check_login_defs {}

1;
