# Copyright (c) International Business Machines Corporation, 2007.  All
# rights reserved. Portions of this file are supplied under license by
# Transitive Limited, which is copyright (c) Transitive Limited, as
# follows:
#
# Copyright (c) Transitive Limited 2007-2008. All rights reserved.
#
# This software program, including any upgrades, patches or programs added
# to it ("Program"), and the Materials relating to it (in particular all IPRs
# in the foregoing) are and shall at all times remain the property of
# Transitive Limited (English company number 04072099), which
# reserves its rights under all applicable law in relation to that property.
# Any use, reproduction or dealings with any part of Transitive's property which
# has not been expressly authorised is explicitly prohibited.  Transitive confirms
# that the Materials, technical data, information and documentation necessary for
# maintaining and correcting the Program are confidential, and that it reserves the
# right to prosecute any use of the Program or the Material made in breach of this
# notice. In relation to this notice, and without limitation:
#
# The term "Materials" shall include all logic, logic diagrams, flowcharts,
# orthographic representations, algorithms, routines, sub-routines, utilities,
# modules, file structures, coding sheets, source codes listings, functional
# specifications, specifications, user operation manuals and all other materials
# and documents necessary to enable a reasonably skilled programmer to maintain,
# amend and enhance the Program and whether in eye-readable or machine-readable
# form; and
#
# "IPRs" shall include all registered and unregistered patents, trademarks,
# service marks, database rights, design rights, copyright, applications for any
# of the foregoing, and any other similar protected rights in any country including
# the right to register or enforce the same.

package Maiken::RO::RHEL;
use base 'Maiken::RO';

=head1 NAME

  Maiken::RO::RHEL - IBM p AVE installation for Red Hat EL

=cut

use strict;
use warnings;

sub RH_REG_FILE        { $_[0]->CONFIG_DIR.'/redhat-activation-detail' }
sub RH_REGISTER_EMAIL  {  'pave@redhat.com'                  }
sub INIT_D             { '/etc/rc.d/init.d' }

sub initialize {
  my ($self) = @_;
  $self->SUPER::initialize;
}


=item B<get_distro_registration_info_interactively>

Called exclusively by get_registration_info_interactively.
Sends an email to RedHat

=cut

sub get_distro_registration_info_interactively {
  my ($self) = @_;
  my @qs =
    (
     [rh_company_name => "Please enter the name of your company:",
     '.+', 'A company name must be provided.', 'ibm_company_name'],
     [rh_registrant_name => "Please enter your name:",
      '.+', 'You must provide your name.', 'ibm_registrant_name'],
     [rh_email_address => "Please enter your email address:",
      $self->EMAIL_REGEXP, 'The email address is compulsory and must have an @ and a valid domain.',
      'ibm_email_address'],
     [rhn_login => "Please enter your RHN login (optional):",
      '([^+%"<^>|\\$()]+)?',
     'RHN login cannot contain any of the following: +%\"<^>|\\$()'],
     [rh_account_number => "Please enter your Red Hat account number (optional):",
      '(\d{6})?', 'Account numbers must contain exactly 6 digits.'],
    );

  my $initial_text = qq{
The Linux/x86 binaries and libraries in the x86 World can be kept up to date
using the Red Hat Network.  To receive this support, you must upgrade your
current Red Hat subscription to support both Linux/x86 and Linux/POWER.
This upgrade is at no additional charge.

To upgrade the subscription on this system, you need to provide some
information that will be emailed to Red Hat Inc.  Once this has been sent,
you will receive an acknowledgement email from Red Hat within approximately
the next 24 hours (1 business day).  You will then receive another email 
confirming that the subscription has been upgraded within approximately 48 
hours (2 business days). 

};

  return $self->register_interactively
    (
     email                       => $self->RH_REGISTER_EMAIL,
     organisation                => 'Red Hat Inc.',
     record_file                 => $self->RH_REG_FILE,
     subject_line_method         => 'rh_subject_line',
     email_body_method           => 'rh_registration_text',
     questions                   => \@qs,
     initial_text_for_first_time => $initial_text,
     compulsory                  => 0,
    );
}

=item B<rh_registration_text>

Returns a string containing the contents of the registration
email to be sent to RH.

=cut

sub rh_registration_text {
  my ($self) = @_;
  my $c = $self->{config};
  my $text = qq{
Date:         $c->{registration_date}

Company:      $c->{rh_company_name}
RHN Login:    $c->{rhn_login}
Account No:   $c->{rh_account_number}

Name:         $c->{rh_registrant_name}
Email:        $c->{rh_email_address}
};
  return $text;
}

=item B<finalize_sw_installation>

As the superclass but run clean_rc_dirs at the beginning.

=cut

sub finalize_sw_installation {
  my ($self) = @_;
  # Let's clear rcX.d
  $self->clean_rc_dirs;
  $self->SUPER::finalize_sw_installation;
}

=item B<clean_rc_dirs>

Cleans the rcX.d dirs

=cut

sub clean_rc_dirs {
  my ($self) = @_;
  # It is probably safer to disable all services using
  # chkconfig and then - to get rid of the reboot, halt and single
  # scripts - unlink *
  my $command = "/sbin/chkconfig";
  # what services do we have?
  my $tempo = $self->run_translated('/bin/ls', '-l', '/sbin/chkconfig');
  my @service_list = split /\n/,$self->run_translated($command, '--list');
  if( $self->{exit} == 0 ) {
    foreach my $line (@service_list) {
      unless( $line =~ /xinetd/ and not $line=~/5:/ ) {
        $line =~ /\s*(.*?)\s+.*/;
        my $service_name = $1;
        $service_name =~ s/://g;
        my $msg = $self->run_translated($command, $service_name, 'off');
        if( $self->{exit} != 0 ) {
          $self->log("WARNING: [_1]'s startup state could not be set.", $service_name);
        }
      }
    }
    # now, unlink * to delete the scripts that chkconfig
    # did not delete (halt, reboot, etc)
    my $scripts_dir = "$self->{config}{sw_root_path}/etc/rc.d";
    for my $i (0..6) {
      my @dir_list = glob("\Q$scripts_dir/rc$i.d\E/*");
      foreach my $file (@dir_list) {
        # rc.local is kept
        unless( $file =~ /S99local$/ ) {
          unlink $file;
        }
      }
    }
    $self->enable_services;
  } else {
    $self->log("WARNING: The list of services could not be retrieved.");
  }
}

=item B<setup_sw>

As the superclass but also copy the /etc/sysconfig/network file into the x86 World at the end.

=cut

sub setup_sw {
  my ($self) = @_;
  $self->SUPER::setup_sw;

  # Copy the native /etc/sysconfig/network file into the x86 World.
  my $network_file = "/etc/sysconfig/network";

  if (-f $network_file) {
    File::Copy::copy($network_file, "$self->{config}{sw_root_path}/$network_file")
     or $self->log("Warning - failed to create [_1]", $network_file);
  }
}

=item B<rh_subject_line>

Returns a string containing the subject line for the 
redhat registration email to be sent to RH.

=cut

sub rh_subject_line {
  my ($self) = @_;

  return 'Red Hat Subscription Upgrade for System p AVE' ;
}


=item B<check_login_defs>

See superclass. The RPM is shadow-utils.

=cut

sub check_login_defs {
  my $self = shift;
  my $rpm = shift;
  if (defined $rpm && ($rpm =~ /shadow-utils/)) {
    $self->modify_login_defs;
  }
}

1;
